/* Copyright 2010 Santiago Ontanon and Ashwin Ram */

package d2.execution.planbase;

import gatech.mmpm.GameState;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

/**
 *
 * @author santi
 */
public class GameStateSimilarity {

	/*
	 *  This function computes a generic similarity metric among game states. It optionally accepts
	 *  a list of weights as a HashMap that maps "feature name" to "weight". If no weights are passed,
	 *  then they are assumed to be uniform.
	 *
	 *  Each feature is identified by a String in the HashMap.
	 *
	 *  It uses generic features to compute the similarity. In particular it uses:
	 *  - histogram of entities in the map
	 *  - histogram of entities in the game state (not map)
	 *  - number of units of each player
	 *  - number of units total in map and game state
	 *  - size of map
	 *  - sensors in the sensor library
	 */
	public static double gameStateSimilarity(GameStateFeatures gs1, GameStateFeatures gs2,HashMap<String,Double> weights) {
		ArrayList<Double> featureValue1 = new ArrayList<Double>(),featureValue2 = new ArrayList<Double>();
		ArrayList<Double> featureWeight = new ArrayList<Double>();

		Set<String> featureNames = new HashSet<String>();
		featureNames.addAll(gs1.getFeatureNames());
		featureNames.addAll(gs2.getFeatureNames());

		for(String f:featureNames) {
			featureValue1.add(gs1.getFeature(f));
			featureValue2.add(gs2.getFeature(f));
			if (weights!=null) {
				featureWeight.add(weights.get(f));
			} else {
				featureWeight.add(1.0);
			}
		}

		return computeSimilarity(featureValue1,featureValue2,featureWeight);
	}

	public static double gameStateSimilarity(GameState rgs1, GameState rgs2,int cycle, String player1, String player2,HashMap<String,Double> weights) {
		GameStateFeatures gs1 = (GameStateFeatures) rgs1.getMetaData("GameStateFeatures");
		if (gs1==null) {
			gs1 = new GameStateFeatures(cycle, rgs1,player1);
			rgs1.addMetaData("GameStateFeatures", gs1);
		}
		GameStateFeatures gs2 = (GameStateFeatures) rgs1.getMetaData("GameStateFeatures");
		if (gs2==null) {
			gs2 = new GameStateFeatures(cycle, rgs2,player2);
			rgs2.addMetaData("GameStateFeatures", gs2);
		}

		ArrayList<Double> featureValue1 = new ArrayList<Double>(),featureValue2 = new ArrayList<Double>();
		ArrayList<Double> featureWeight = new ArrayList<Double>();

		Set<String> featureNames = new HashSet<String>();
		featureNames.addAll(gs1.getFeatureNames());
		featureNames.addAll(gs2.getFeatureNames());

		for(String f:featureNames) {
			featureValue1.add(gs1.getFeature(f));
			featureValue2.add(gs2.getFeature(f));
			if (weights!=null) {
				featureWeight.add(weights.get(f));
			} else {
				featureWeight.add(1.0);
			}
		}

		return computeSimilarity(featureValue1,featureValue2,featureWeight);

	}


	static double computeSimilarity(ArrayList<Double> features1,ArrayList<Double> features2,ArrayList<Double> weights) {
		double accum = 0;
		double accumWeight = 0;

		for(int i=0;i<features1.size();i++) {
			double d = 1 - Math.abs(features1.get(i)-features2.get(i))/(Math.max(features1.get(i), features2.get(i))+1);
			accum += d*weights.get(i);
			accumWeight+=weights.get(i);
		}

		if (accumWeight>0) {
			return accum/accumWeight;
		} else {
			return 0;
		}
	}
}
