/* Copyright 2010 Santiago Ontanon and Ashwin Ram */

package d2.execution.planexecution;

import d2.core.D2;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.LinkedList;
import java.util.List;

import org.jdom.Element;

import d2.execution.adaptation.PlanAdaptation;
import d2.execution.planbase.PlanBase;
import d2.execution.planner.Planner;
import d2.plans.ActionPlan;
import d2.plans.FailureTransition;
import d2.plans.GoalPlan;
import d2.plans.PetriNetElement;
import d2.plans.PetriNetPlan;
import d2.plans.Plan;
import d2.plans.PlanPreconditionFailedState;
import d2.plans.PlanState;
import d2.plans.PreFailureTransition;
import d2.plans.PreTransition;
import d2.plans.State;
import d2.plans.SuccessTransition;
import d2.plans.Transition;
import gatech.mmpm.Action;
import gatech.mmpm.GameState;

import gatech.mmpm.sensor.Sensor;
import gatech.mmpm.util.Pair;
import gatech.mmpm.util.XMLWriter;

public abstract class PlanExecution {

    protected PlanAdaptation m_parameterAdapter = null;
    PlanAdaptation m_planAdapter = null;
    D2 m_d2 = null;
    String currentPlayer = null;
    public static int DEBUG = 1;
    public static int EXECUTION_DEBUG = 1;
    public static final int PLAN_FINISHED = 1;
    public static final int PLAN_ADVANCED = 2;
    protected List<Pair<Plan, Plan>> plansThatCouldntStart = new LinkedList<Pair<Plan, Plan>>();
    
    // This list keeps track of all the units who are currently executing actions, which have yet 
    // to succeed or fail:
    protected List<String> unitsInUse = new LinkedList<String>();

    public void setD2(D2 d2) {
        m_d2 = d2;
        m_parameterAdapter.setD2(d2);
        m_planAdapter.setD2(d2);
    }

    public void setPlanAdaptation(PlanAdaptation pa) {
        m_planAdapter = pa;
    }

    public void setParameterAdaptation(PlanAdaptation ppa) {
        m_parameterAdapter = ppa;
    }

    public PlanExecution(String player, PlanAdaptation parameterAdapter, PlanAdaptation planAdapter, D2 d2) {
        currentPlayer = player;
        m_parameterAdapter = parameterAdapter;
        m_planAdapter = planAdapter;
        m_d2 = d2;
        if (m_parameterAdapter!=null) m_parameterAdapter.setD2(m_d2);
        if (m_planAdapter!=null) m_planAdapter.setD2(m_d2);
    }

    public List<String> getUnitsInUse() {
        return unitsInUse;
    }

    public void setPlayer(String player) {
        currentPlayer = player;
        m_parameterAdapter.setPlayer(player);
        m_planAdapter.setPlayer(player);
    }

    //returns true if plan is executed completely
    //else returns false
    public abstract boolean execute(Planner planner, List<Action> a_actionsToExecute, int cycle, GameState a_gs, String player);

    public abstract void saveToXML(XMLWriter w);

    public int executePlan(Plan topLevelPlan, Plan a_plan, List<Plan> a_plansToExecute, List<Plan> a_plansToRemove, int cycle, GameState a_gs, String player) {
        int retval = PLAN_FINISHED;
        if (EXECUTION_DEBUG >= 2) {
            System.out.println("executePlan: " + a_plan);
        }

        if (a_plan instanceof ActionPlan) {
            retval = executeAction((ActionPlan) a_plan, a_plansToExecute, a_plansToRemove, cycle, a_gs, player);
        } else if (a_plan instanceof GoalPlan) {
            retval = executeGoal(topLevelPlan, (GoalPlan) a_plan, a_plansToExecute, a_plansToRemove, cycle, a_gs, player);
        } else if (a_plan instanceof PetriNetPlan) {
            retval = executePetriNetFAST(topLevelPlan, (PetriNetPlan) a_plan, a_plansToExecute, a_plansToRemove, cycle, a_gs, player);
        }

        if ((retval & PLAN_FINISHED) == PLAN_FINISHED) {
            a_plansToRemove.add(a_plan);
        }
        return retval;
    }

    public int executeAction(ActionPlan a_plan, List<Plan> a_plansToExecute, List<Plan> a_plansToRemove, int cycle, GameState a_gs, String player) {
        if (EXECUTION_DEBUG >= 2) {
            System.out.println("executeAction");
        }

        a_plansToExecute.add(a_plan);
        return PLAN_FINISHED;
    }

    public int executeGoal(Plan topLevelPlan, GoalPlan a_plan, List<Plan> a_plansToExecute, List<Plan> a_plansToRemove, int cycle, GameState a_gs, String player) {
        Plan expandedPlan = a_plan.getExpandedGoalPlan();

        if (EXECUTION_DEBUG >= 2) {
            System.out.println("executeGoal");
        }

        if (expandedPlan == null) {
            return 0;
        } else {
            int status = executePlan(topLevelPlan, expandedPlan, a_plansToExecute, a_plansToRemove, cycle, a_gs, player);
            if ((status & PLAN_FINISHED) == PLAN_FINISHED) {
                expandedPlan = null;
            }

            int retval = 0;
            if ((status & PLAN_ADVANCED) == PLAN_ADVANCED) {
                retval |= PLAN_ADVANCED;
            }
            if (a_plan.checkSuccessCondition(cycle, a_gs, player)) {
                retval |= PLAN_FINISHED;
            }
            return retval;
        }
    }

    public int executePetriNetFAST(Plan topLevelPlan, PetriNetPlan a_plan, List<Plan> a_plansToExecute, List<Plan> a_plansToRemove, int cycle, GameState a_gs, String player) {
        int planOver = PLAN_FINISHED;
//		List<Pair<State,Integer>> tokensToAdd = new LinkedList<Pair<State,Integer>>();

        List<Integer> markedStates = new LinkedList<Integer>();

        if (EXECUTION_DEBUG >= 2) {
            System.out.println("executePetriNetFAST");
        }

        for (int i = 0; i < a_plan.getPetriNetElements().size(); i++) {
            PetriNetElement p = a_plan.getPetriNetElement(i);

            if (p instanceof State) {
                State s = (State) p;
                for (Pair<Integer, Transition> t_p : s.getNextTransitions()) {
                    if (s.getCurrentNumberOfTokens() >= t_p._a && !markedStates.contains(i)) {
                        planOver = 0;
                        markedStates.add(i);
                    }
                }
            }
        }

//		if markedStates is empty, return the planOver
        if (markedStates.size() == 0) {
            return PLAN_FINISHED;
        }

//		Go thru the list again, and this time hit all the markedStates and move tokens till you get an action
//		or are stuck at a condition
        if (EXECUTION_DEBUG >= 2) {
            System.out.println("executePetriNetFAST: Found " + markedStates.size() + " marked states!");
        }

        for (int i = 0; i < a_plan.getPetriNetElements().size(); i++) {
            if (markedStates.contains(i)) {
                State s = (State) a_plan.getPetriNetElement(i);

                // TOOD: [santi]: Jai, are you sure this works. How can you assume that the
                //				  executeStateChainTillAction method does not change the value of "markedStates"?
                //	call to function that executes the chain till an action
                if (executeStateChainTillAction(topLevelPlan, a_plan, s, a_plansToExecute, cycle, a_gs, player)) {
                    planOver |= PLAN_ADVANCED;
                }
            }
        }

        return planOver;
    }

    private boolean executeStateChainTillAction(Plan topLevelPlan, PetriNetPlan a_plan, State s, List<Plan> a_plansToExecute, int cycle, GameState a_gs, String player) {
        boolean tokensMoving = false;
        Plan adaptedPlan = null, originalPlan = null;

        if (EXECUTION_DEBUG >= 2) {
            System.out.println("\texecutePetriNetFAST: executeStateChainTillAction: in state " + s.getElementID() + " tokens: " + s.getCurrentNumberOfTokens());
        }

        for (Pair<Integer, Transition> t_p : s.getNextTransitions()) {
            if (s.getCurrentNumberOfTokens() >= t_p._a) {

//                System.out.println("Transition type: " + t_p._b.getClass().getSimpleName());

                originalPlan = t_p._b.getPlan();
                adaptedPlan = null;
                if (t_p._b instanceof PreTransition) {
                    // If we are about to start a subplan, adapt it!
                    if (EXECUTION_DEBUG >= 1) {
                        System.out.println("About to Adapt: " + originalPlan);
                        System.out.println("Units in use are: " + unitsInUse);
                    }
                    adaptedPlan = m_parameterAdapter.adapt(originalPlan, originalPlan.getOriginalGameState(), originalPlan.getOriginalPlayer(), null, cycle, a_gs, player, null,unitsInUse);
                    if (adaptedPlan == null) {
//						t_p.m_b.setPlan(adapted);
                        if (EXECUTION_DEBUG >= 1) {
                            System.out.println("executeStateChainTillAction: stopped, since the plan cannot be adapted.");
                        }
//                        return tokensMoving;
                    } else {
                        if (EXECUTION_DEBUG >= 1) System.out.println("Action successfully adapted");
                    }
                }

                // If adaptedPlan != null, that means that the transition is a PRETRANSITION, and the Sensor is thus
                // the precondition of the plan
                if ((adaptedPlan != null && adaptedPlan.checkPreCondition(cycle, a_gs, player))
                    || (adaptedPlan == null && t_p._b instanceof SuccessTransition && t_p._b.getPlan().checkSuccessCondition(cycle, a_gs, player))
                    || (adaptedPlan == null && t_p._b instanceof FailureTransition && t_p._b.getPlan().checkFailureCondition(cycle, a_gs, player))
                    || (adaptedPlan == null && t_p._b instanceof PreFailureTransition && t_p._b.getPlan().checkPreFailureCondition(cycle, a_gs, player))
                    || (adaptedPlan == null && t_p._b.getPlan() == null && (Float) t_p._b.getCondition().evaluate(cycle, a_gs, player, a_plan.getContext(cycle, a_gs, player)) >= Sensor.BOOLEAN_TRUE_THRESHOLD)) {

                    if (adaptedPlan != null) {
                        // Replace this plan in the planstate, pretransition, successtransition, and failuretransition
                        a_plan.replacePlan(t_p._b.getPlan(), adaptedPlan);
                        if (EXECUTION_DEBUG >= 1) {
                            System.out.println("originalPlan: " + originalPlan);
                        }
                        if (EXECUTION_DEBUG >= 1) {
                            System.out.println("adaptedPlan: " + adaptedPlan);
                        }

                        if (adaptedPlan instanceof ActionPlan) {
                            unitsInUse.add(((ActionPlan)adaptedPlan).getAction().getEntityID());
                        }
                    }

                    if (t_p._b instanceof SuccessTransition ||
                        t_p._b instanceof FailureTransition) {
                        Plan p = t_p._b.getPlan();
                        if (p!=null && p instanceof ActionPlan) {
                            unitsInUse.remove(((ActionPlan)p).getAction().getEntityID());
                        }
                    }

//				 	System.out.println("Condition " + t_p.m_b.getCondition() + " satisfied of a " + t_p.m_b.getClass().getSimpleName() + " [cycle: " + cycle + "]");

//					If it's a precondition transition, execute the plan:
//					and its an action (or subgoalplan, or petrinetplan...whatever), so well...here goes the action...

                    if (EXECUTION_DEBUG >= 1) {
                        if (t_p._b instanceof PreFailureTransition
                            || t_p._b instanceof PreTransition
                            || t_p._b instanceof SuccessTransition
                            || t_p._b instanceof FailureTransition) {
                            System.out.println(t_p._b.getClass().getSimpleName() + " fired for plan " + t_p._b.getPlan());
                            System.out.println("With context: " + t_p._b.getPlan().getContext(cycle, a_gs, player));
                            if (t_p._b.getPlan() instanceof ActionPlan) {
                                System.out.println("Unit: " + a_gs.getEntity(((ActionPlan)t_p._b.getPlan()).getAction().getEntityID()));
                            }
                            System.out.println(t_p._b.getCondition());

                            // clear the potential fields for the action just executed:
                            if (t_p._b instanceof SuccessTransition
                                    || t_p._b instanceof FailureTransition) {
                                Plan p = t_p._b.getPlan();
                                GameState gs = p.getOriginalGameState();
                                gs.clearMetadata();
                            }

                            if (EXECUTION_DEBUG >= 2) {
                                System.out.println(" /- cycle " + cycle + " --------------------\\ ");
                                Planner.printPlanNice(topLevelPlan);
                                System.out.println(" \\-------------------------/ ");
                            }

                        }
                    }

                    if (t_p._b instanceof PreTransition) {
                        if (EXECUTION_DEBUG >= 2) {
                            System.out.println("\texecutePetriNetFAST:Fired Transition" + t_p._b.getElementID() + " removed tokens from state " + s.getElementID() + " -> " + s.getCurrentNumberOfTokens() + " (condition: " + t_p._b.getCondition() + ", plan: " + t_p._b.getPlan() + ")");
                        }

                        if (t_p._b.getPlan() != null) {
                            a_plansToExecute.add(t_p._b.getPlan());
                        } else {
                            System.err.println("Plan null in petri net element! -> " + t_p._b.getElementID());
                        }
                        tokensMoving = true;
                        s.setCurrentNumberOfTokens(s.getCurrentNumberOfTokens() - t_p._a);
//                        System.out.println("Tokens in state " + s.getElementID() + " now " + s.getCurrentNumberOfTokens() + " (down " + t_p._a + ")");
                        for (Pair<Integer, State> s_p : t_p._b.getNextStates()) {
                            s_p._b.setCurrentNumberOfTokens(s_p._b.getCurrentNumberOfTokens() + s_p._a);
//                            System.out.println("Tokens in state " + s_p._b.getElementID() + " now " + s_p._b.getCurrentNumberOfTokens() + " (up " + s_p._a + ")");
                            if (EXECUTION_DEBUG >= 2) {
                                System.out.println("\texecutePetriNetFAST:Added " + t_p._a + " token to state " + s_p._b.getElementID());
                            }
                        }
                        return tokensMoving;
                    } else {
                        if (EXECUTION_DEBUG >= 2) {
                            System.out.println("\texecutePetriNetFAST:Fired Transition" + t_p._b.getElementID() + " removed tokens from state " + s.getElementID() + " -> " + s.getCurrentNumberOfTokens() + " (condition: " + t_p._b.getCondition() + ", plan: " + t_p._b.getPlan() + ")");
                        }

                        // Remove the tokens from the preceding state:
                        tokensMoving = true;
                        s.setCurrentNumberOfTokens(s.getCurrentNumberOfTokens() - t_p._a);
//                        System.out.println("Tokens in state " + s.getElementID() + " now " + s.getCurrentNumberOfTokens() + " (down " + t_p._a + ")");
                        for (Pair<Integer, State> s_p : t_p._b.getNextStates()) {
                            s_p._b.setCurrentNumberOfTokens(s_p._b.getCurrentNumberOfTokens() + s_p._a);
//                            System.out.println("Tokens in state " + s_p._b.getElementID() + " now " + s_p._b.getCurrentNumberOfTokens() + " (up " + s_p._a + ")");
                            executeStateChainTillAction(topLevelPlan, a_plan, s_p._b, a_plansToExecute, cycle, a_gs, player);
                            if (s_p._b instanceof PlanPreconditionFailedState
                                    && s_p._b.getNextTransitions().size() == 0) {
                                // The preconditions are not satisfied, and the action has not been adapted before:
                                plansThatCouldntStart.add(new Pair<Plan, Plan>(s_p._b.getPlan(), a_plan));
                            }
                        }
                    }
                } else {
                    if (EXECUTION_DEBUG >= 2) {
                        if (adaptedPlan == null) {
                            System.out.println("AdaptedPlan = null");
                            System.out.println("OriginalPlan = " + originalPlan);
                            if (t_p._b instanceof SuccessTransition) {
                                Sensor sensor = t_p._b.getCondition();
                                System.out.println("executePetriNetFAST: waiting for SuccessCondition " + sensor + " [cycle: " + cycle + "]");
                                System.out.println("Which evaluates to: " + t_p._b.getPlan().checkSuccessCondition(cycle, a_gs, player));
                            } else if (t_p._b instanceof FailureTransition) {
                                Sensor sensor = t_p._b.getCondition();
                                System.out.println("executePetriNetFAST: waiting for FailureCondition " + sensor + " [cycle: " + cycle + "]");
                                System.out.println("Which evaluates to: " + t_p._b.getPlan().checkFailureCondition(cycle, a_gs, player));
                            } else {
                                Sensor sensor = t_p._b.getCondition();
                                System.out.println("executePetriNetFAST: waiting for Sensor " + sensor + " [cycle: " + cycle + "]");
                                System.out.println("Which evaluates to: " + sensor.evaluate(cycle, a_gs, player));
                            }
                        } else {
                            Sensor sensor = adaptedPlan.getPreCondition();
                            System.out.println("executePetriNetFAST: waiting for PreCondition " + sensor + ", plan: " + adaptedPlan + " [cycle: " + cycle + "]");
                            System.out.println("Which evaluates to: " + adaptedPlan.checkPreCondition(cycle, a_gs, player));
                        }
                    }
                }
            }
        }
        return tokensMoving;
    }

    public void removeStillUsedUnitIDs(Plan p) {
        if (p instanceof ActionPlan) {
            unitsInUse.remove(((ActionPlan)p).getAction().getEntityID());
        } else if (p instanceof GoalPlan) {
            if (((GoalPlan)p).getExpandedGoalPlan()!=null) {
                removeStillUsedUnitIDs(((GoalPlan)p).getExpandedGoalPlan());
            }
        } else if (p instanceof PetriNetPlan) {
            for(PlanState s:((PetriNetPlan)p).getAllPlanStates()) {
                if (s.getCurrentNumberOfTokens()>0) {
                    removeStillUsedUnitIDs(s.getPlan());
                }
            }
        }
    }

    public static PlanExecution loadFromXML(Element xml, String a_player, D2 d2) {
        Class<?> c;
        try {

            c = Class.forName(xml.getChildText("type"));
            Method m = c.getMethod("loadFromXMLInternal", new Class[]{Element.class, String.class, D2.class});
            return (PlanExecution) m.invoke(c, xml, a_player, d2);
        } catch (ClassNotFoundException e) {
            e.printStackTrace();
        } catch (SecurityException e) {
            e.printStackTrace();
        } catch (NoSuchMethodException e) {
            e.printStackTrace();
        } catch (IllegalArgumentException e) {
            e.printStackTrace();
        } catch (IllegalAccessException e) {
            e.printStackTrace();
        } catch (InvocationTargetException e) {
            e.printStackTrace();
        }
        return null;
    }

    public static PlanExecution loadFromXMLInternal(Element xml, String a_player, D2 d2) {
        return null;
    }
}
