/* Copyright 2010 Santiago Ontanon and Ashwin Ram */

package d2.learn.planlearning.goalrec;

import d2.core.D2;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import d2.execution.adaptation.ExpandConditionMatcher;
import d2.execution.adaptation.ExpandConditionMatcher.Match;
import d2.execution.planbase.PlanBase;
import d2.execution.planbase.PlanBaseEntry;
import d2.learn.planlearning.PlanLearning;
import d2.plans.ActionPlan;
import d2.plans.DummyState;
import d2.plans.DummyTransition;
import d2.plans.FailureTransition;
import d2.plans.PetriNetElement;
import d2.plans.PetriNetPlan;
import d2.plans.Plan;
import d2.plans.PlanPreconditionFailedState;
import d2.plans.PlanState;
import d2.plans.PreTransition;
import d2.plans.State;
import d2.plans.SuccessTransition;
import d2.plans.Transition;
import d2.worldmodel.ConditionMatcher;

import gatech.mmpm.Action;
import gatech.mmpm.Entry;
import gatech.mmpm.GameState;
import gatech.mmpm.Trace;

import gatech.mmpm.sensor.Sensor;
import gatech.mmpm.sensor.composite.AndCondition;
import gatech.mmpm.sensor.constant.False;
import gatech.mmpm.sensor.constant.True;

/**
 * @author kane
 * 
 * Learns plan hierarchy given the list of GameStates and actions.
 */
public class DependencyPlanLearning extends PlanLearning {
	/** epsilon for the difference between GameStateDeltas. */
	private static double EPSILON = 0.001;

	/** the player whose plan is being learned */
	private String player;

        private ConditionMatcher m_cm;

	private int DEBUG = 0;

	public DependencyPlanLearning(D2 a_d2, String domainName, ConditionMatcher cm) {
		super(a_d2, domainName);
                m_cm = cm;
	}

	public void learnFromTrace(Trace t, String playerName) {
		player = playerName;
		if (DEBUG >= 1) {
			System.out.println("DependencyPlanLearning.learnFromTrace: ");
		}
		learn(t.getEntries(), playerName);

		System.out.println("actions: " + m_d2.getPlanBase().getAllActions().size());
		System.out.println("plans: " + m_d2.getPlanBase().getPlans().size());
	}

	/**
	 * Learns the dependencies and tree relation of a gameLog.
	 * 
	 * @param gameLog
	 *            the log of GameStates and Actions.
	 * @param rulesFile
	 *            the game specific rules.
	 * @return
	 */
	private List<GameTreeNode> learn(List<Entry> gameLog, String playerID) {
		if (DEBUG >= 1) {
			System.out.println("DependencyPlanLearning.learn: ");
		}
		// extract the game log into a list of nodes
		ArrayList<GameTreeNode> nodes = new ArrayList<GameTreeNode>();
		for (Entry state : gameLog) {

			// find the action that corresponds to the player
			Action act = findPlayerAction(state);
			// if the state didn't have an action from the player, skip it
			if (null == act) {
				continue;
			}
			PetriNetPlan pnp = createActionPNP(act, state.getGameState());

			nodes.add(new GameTreeNode(state.getGameState(), pnp));
		}
		discoverDependencies(nodes);

		buildDependencyTrees(nodes);

		addPlansToPlanBase(nodes, playerID);

		// for (Plan plan : planDB.getPlans()) {
		// PlanVisualizer pv = new PlanVisualizer(plan, 800, 600, nodeNum,
		// null);
		// }

		return nodes;
	}

	/**
	 * Find the action in the state for the given player.
	 * 
	 * @param state
	 * @return
	 */
	private Action findPlayerAction(Entry state) {
		Action act = null;
		for (Action a : state.getActions()) {
			if (a.getPlayerID() != null && a.getPlayerID().equals(player)) {
				act = a;
				break;
			}
		}
		return act;
	}

	/**
	 * Add each node in the tree as a progressively more detailed plan to the
	 * planBase.
	 * 
	 * @param nodes
	 */
	private void addPlansToPlanBase(List<GameTreeNode> nodes, String playerName) {
		for (GameTreeNode node : nodes) {
			addPlansToPlanBase(node.getChildren(), playerName);

			GameState earliest = getEarliestChildState(node);
			Sensor g = new DeltaCondition(earliest, node.getGs(), d2.core.Config.getDomain()
					.getName());
			PetriNetPlan pnp = node.getPlan();
			PlanBaseEntry pbe = new PlanBaseEntry(pnp, g, earliest, playerName);
			PetriNetElement petrinet_head = pnp.getAllElements().get(0);
			HashMap<String, Integer> tokensInStates = new HashMap<String, Integer>();
			tokensInStates.put(petrinet_head.getElementID(), 1);
			pbe.setEntryPoint(tokensInStates);

			m_d2.getPlanBase().addPlan(pbe);
		}
	}

	/**
	 * Find the GameState for the leftmost child of this node, or the GameState
	 * for this node if it has no children.
	 * 
	 * @param node
	 *            the parent node.
	 * @return the earliest GameState
	 */
	private GameState getEarliestChildState(GameTreeNode node) {
		if (node.getChildren().size() == 0) {
			return node.getGs();
		} else {
			return getEarliestChildState(node.getChildren().get(0));
		}
	}

	/**
	 * find dependencies between nodes, and condense them until there are no
	 * more to be condensed.
	 * 
	 * @param nodes
	 */
	private void buildDependencyTrees(ArrayList<GameTreeNode> nodes) {
		if (DEBUG >= 1) {
			System.out.println("DependencyPlanLearning.buildDependencyTrees: ");
		}
		while (true) {
			double minDelta = findMinDelta(nodes);

			if (minDelta > 1.0) {
				break;
			}

			List<GameTreeNode> toCondense = findToCondense(minDelta, nodes);

			GameTreeNode lastGameTreeNode = findLatestGameTreeNode(toCondense, nodes);
			List<DependencyNode> dep = new ArrayList<DependencyNode>();

			// combine toCondense. Always at least two nodes
			PetriNetPlan pnp = null;
			GameTreeNode first = null;
			for (GameTreeNode gtn : toCondense) {
				if (null == first) {
					first = gtn;
				} else {
					pnp = createCombinedPNP(first.getPlan(), gtn.getPlan());
				}
			}
			

			GameTreeNode newNode = new GameTreeNode(lastGameTreeNode.getGs(), toCondense, dep, pnp);

			updateNewNodeDependencies(toCondense, nodes, newNode);

			nodes.set(nodes.indexOf(lastGameTreeNode), newNode);
			nodes.removeAll(toCondense);

			updateOtherDependencies(nodes, newNode);

		}
	}

	/**
	 * Update nodes that depended on the children of the newNode to depend on
	 * the newNode
	 * 
	 * @param nodes
	 * @param newNode
	 */
	private void updateOtherDependencies(List<GameTreeNode> nodes, GameTreeNode newNode) {
		List<GameTreeNode> children = newNode.getChildren();
		for (GameTreeNode n : nodes) {
			if (n.equals(newNode)) {
				continue;
			}
			List<DependencyNode> toRemove = new ArrayList<DependencyNode>();
			List<DependencyNode> toAdd = new ArrayList<DependencyNode>();
			for (DependencyNode dn : n.getDependencies()) {
				if (children.contains(dn.getDepended())) {
					toRemove.add(dn);
					GameStateDelta delta = new GameStateDelta(newNode.getGs(), dn.getDepending()
							.getGs());
					DependencyNode newDN = new DependencyNode(delta, newNode, dn.getDepending());
					toAdd.add(newDN);
				}
			}
			n.getDependencies().removeAll(toRemove);
			n.getDependencies().addAll(toAdd);
		}
	}

	/**
	 * Find the GameState corresponding the latest occurrence of subset in
	 * superset.
	 * 
	 * @param subset
	 * @param superset
	 * @return
	 */
	private GameTreeNode findLatestGameTreeNode(List<GameTreeNode> subset,
			List<GameTreeNode> superset) {
		int index = 0;
		for (GameTreeNode n : subset) {
			int curIndex = superset.indexOf(n);
			if (curIndex > index) {
				index = curIndex;
			}
		}

		return superset.get(index);

	}

	private void updateNewNodeDependencies(List<GameTreeNode> toCondense, List<GameTreeNode> nodes,
			GameTreeNode newNode) {

		// find and add Dependencies the newNode has.
		for (GameTreeNode n : toCondense) {
			for (DependencyNode d : n.getDependencies()) {
				GameTreeNode gtn = d.getDepended();
				if (toCondense.contains(gtn)) {
					continue;
				}
				GameStateDelta delta = new GameStateDelta(gtn.getGs(), newNode.getGs());
				DependencyNode newDN = new DependencyNode(delta, gtn, newNode);
				newNode.getDependencies().add(newDN);
			}
		}

	}

	/**
	 * find GameTreeNodes that need to be condensed to each other.
	 * 
	 * @param minDelta
	 * @return
	 */
	private List<GameTreeNode> findToCondense(double minDelta, ArrayList<GameTreeNode> nodes) {
		List<GameTreeNode> toCondense = new ArrayList<GameTreeNode>();
		List<DependencyNode> minNodes = new ArrayList<DependencyNode>();

		// find minimum dependencies
		for (GameTreeNode node : nodes) {
			for (DependencyNode dep : node.getDependencies()) {
				if (Math.abs(dep.getDelta().getDelta() - minDelta) < EPSILON) {
					minNodes.add(dep);
				}
			}
		}

		// seed list with first node
		DependencyNode node = minNodes.get(0);
		toCondense.add(node.getDepended());
		toCondense.add(node.getDepending());
		minNodes.remove(0);

		// check that any other nodes we add will condense to the same node
		for (DependencyNode dn : minNodes) {
			if (toCondense.contains(dn.getDepended()) && !toCondense.contains(dn.getDepending())) {
				toCondense.add(dn.getDepending());
			}
			if (!toCondense.contains(dn.getDepended()) && toCondense.contains(dn.getDepending())) {
				toCondense.add(dn.getDepended());
			}
		}

		// if more nodes had the same delta, they condense to a separate node
		// which we will get on the next pass

		return toCondense;
	}

	/**
	 * find the minimum delta between related nodes.
	 * 
	 * @param nodes
	 */
	private double findMinDelta(ArrayList<GameTreeNode> nodes) {
		double minDelta = 2.0; // all deltas are between 0.0 and 1.0, so
		// this is higher than any.
		for (GameTreeNode node : nodes) {
			for (DependencyNode depNode : node.getDependencies()) {
				if (depNode.getDelta().getDelta() < minDelta) {
					minDelta = depNode.getDelta().getDelta();
				}
			}
		}
		return minDelta;
	}

	/**
	 * discover the dependencies between nodes
	 * 
	 * @param nodes
	 * @param matcher
	 */
	private void discoverDependencies(ArrayList<GameTreeNode> nodes) {
		if (DEBUG >= 1) {
			System.out.println("DependencyPlanLearning.discoverDependencies: ");
		}
		ExpandConditionMatcher matcher = new ExpandConditionMatcher(m_cm);

		for (int i = 1; i < nodes.size(); i++) {
			// start with node and work backwards to get dependencies
			Plan second = nodes.get(i).getPlan();
			List<Sensor> allPre = new ArrayList<Sensor>();
			ExpandConditionMatcher.expandCondition(second.getPreCondition(), allPre);
			for (int j = i - 1; j >= 0; j--) {
				// check most recent actions and work backwards
				Plan first = nodes.get(j).getPlan();

				List<Match> matches = matcher.matchSuccessWithPreconditionsDetailed(first, 0, first.getOriginalGameState(), first.getOriginalPlayer(), 
                                                                                    second, 0,second.getOriginalGameState(), second.getOriginalPlayer());
				for (Match match : matches) {
					if (allPre.contains(match.preCondition)) {
						// found a matching condition

						// remove the Sensor (don't want more nodes than
						// necessary)
						allPre.remove(match.preCondition);

						// add a dependency
						GameState firstNode = nodes.get(j).getGs();
						GameState secondNode = nodes.get(i).getGs();
						GameStateDelta delta = new GameStateDelta(firstNode, secondNode);
						delta.findDelta(player);
						DependencyNode depNode = new DependencyNode(delta, nodes.get(j), nodes
								.get(i));
						nodes.get(i).getDependencies().add(depNode);
					}
				}

				if (allPre.size() == 0) {
					// no point in checking further, and we don't want extra
					// nodes.
					break;
				}
			}
		}
	}

	/**
	 * Handles all actions within an entry.
	 * 
	 * @param act
	 * @param gs
	 */
	private PetriNetPlan createActionPNP(Action act, GameState gs) {

		PetriNetPlan pnp = new PetriNetPlan();
		Sensor pre_condition = new True();
		DummyState head_ds = new DummyState();
		
		pnp.addPetriNetElement(head_ds); // we're done with head_ds
		DummyState next_ds = new DummyState();

		// now handle the actions block
		ActionPlan ap = new ActionPlan();
		ap.setOriginalGameState(gs);
        ap.setOriginalPlayer(player);

		ap.setAction(act);

		// havta and the pre_condition with the action's pre_condition
		AndCondition act_pre_condition = new AndCondition();
		act_pre_condition.addChild(pre_condition);
		act_pre_condition.addChild(act.getPreCondition());

		// now, create the plan state for that action plan
		PlanState ps = new PlanState();
		ps.setPlan(ap);

		// now the transitions

		// first we create the pre transition from the token dummy state
		// (token_ds)
		PreTransition from_token_to_plan_state = new PreTransition(ap);
		from_token_to_plan_state.addNextState(1, ps);
		head_ds.addNextTransitions(1, from_token_to_plan_state);

		// next we create the success transition from plan state to the next
		// dummy state (next_ds)
		SuccessTransition from_plan_state_to_next_dummy = new SuccessTransition(ap);
		from_plan_state_to_next_dummy.addNextState(1, next_ds);
		ps.addNextTransitions(1, from_plan_state_to_next_dummy);

		// finally we deal with the failure transition, which requires that
		// we create a new "failure" dummy state
		// But we only do it if the action has a non-empty Failure Condition
		if (!(ap.getFailureCondition() instanceof False)) {
			DummyState fs = new DummyState("FAILURESTATE");
			FailureTransition from_plan_state_to_failure_state = new FailureTransition(ap);
			from_plan_state_to_failure_state.addNextState(1, fs);
			ps.addNextTransitions(1, from_plan_state_to_failure_state);
			pnp.addPetriNetElement(from_plan_state_to_failure_state); // failure_transition
			pnp.addPetriNetElement(fs); // failure state
		}

		// If there is a preFailureCondition, add it:
		if (!(ap.getPreFailureCondition() instanceof False)) {
			PlanPreconditionFailedState fs = new PlanPreconditionFailedState();
			fs.setPlan(ap);
			Transition from_dummy_state_to_failure_state = new DummyTransition();
			from_dummy_state_to_failure_state.setCondition((ap).getPreFailureCondition());
			from_dummy_state_to_failure_state.addNextState(1, fs);
			head_ds.addNextTransitions(1, from_dummy_state_to_failure_state);
			pnp.addPetriNetElement(from_dummy_state_to_failure_state); // failure_transition
			pnp.addPetriNetElement(fs); // failure state
		}

		// add all the action specific elements to pnp
		pnp.addPetriNetElement(from_token_to_plan_state); // pre_transition
		pnp.addPetriNetElement(ps); // plan state
		pnp.addPetriNetElement(from_plan_state_to_next_dummy); // success_transition
		pnp.addPetriNetElement(next_ds);
		
		for (PetriNetElement ele : pnp.getAllElements()) {
			System.out.println(ele + " : " + ele.getElementID());
		}
		System.out.println("===");
		
		return pnp;
	}
	
	/**
	 * Combines two PetriNetPlans 
	 * @param first 
	 * @param second
	 */
	private PetriNetPlan createCombinedPNP(PetriNetPlan first, PetriNetPlan second) {

		PetriNetPlan pnp = new PetriNetPlan();

		State firstTail = (State) first.getPetriNetElement(first.getPetriNetElements().size() - 2);
		State secondHead = (State) second.getPetriNetElement(0);
		
		firstTail.getNextTransitions().clear();
		DummyTransition from_first_to_second = new DummyTransition();
		from_first_to_second.addNextState(1, secondHead);
		firstTail.addNextTransitions(1, from_first_to_second);
		
		for (PetriNetElement pne : first.getPetriNetElements()) {
			pnp.addPetriNetElement(pne);
		}
		
		//replace the first's success condition
		pnp.getAllElements().remove(pnp.getAllElements().size() - 1);
		pnp.addPetriNetElement(from_first_to_second);
		
		for (PetriNetElement pne : second.getPetriNetElements()) {
			pnp.addPetriNetElement(pne);
		}

		return pnp;
	}
}
